import 'dart:async';

import 'package:bloc_stream/bloc_stream.dart';
import 'package:riverpod/riverpod.dart';
import 'package:meta/meta.dart';

// ignore: implementation_imports
import 'package:riverpod/src/framework.dart';

import 'builders.dart';

class _BlocStreamProviderState<Bloc extends BlocStream<Value>, Value>
    extends ProviderStateBase<Bloc, Value> {
  StreamSubscription<Value>? subscription;

  @override
  void valueChanged({Bloc? previous}) {
    if (createdValue == previous) {
      return;
    }
    subscription?.cancel();
    _listener(createdValue.value);
    subscription = createdValue.skip(1).listen(_listener);
  }

  void _listener(Value value) {
    exposedValue = value;
  }

  @override
  void dispose() {
    subscription?.cancel();
    super.dispose();
  }
}

mixin _BlocStreamProviderMixin<Bloc extends BlocStream<Value>, Value>
    on RootProvider<Bloc, Value> {
  ProviderBase<Bloc, Bloc> get bloc;

  Override overrideWithValue(Bloc value) {
    return ProviderOverride(
      ValueProvider<Object?, Bloc>((ref) => value, value),
      bloc,
    );
  }
}

// ignore: subtype_of_sealed_class
@sealed
class _BlocStreamProvider<Bloc extends BlocStream<Object?>>
    extends Provider<Bloc> {
  _BlocStreamProvider(
    Create<Bloc, ProviderReference> create, {
    String? name,
  }) : super((ref) {
          final bloc = create(ref);
          ref.onDispose(bloc.close);
          return bloc;
        }, name: name == null ? null : '$name.bloc');
}

// ignore: subtype_of_sealed_class
@sealed
class BlocStreamProvider<Bloc extends BlocStream<Value>, Value>
    extends AlwaysAliveProviderBase<Bloc, Value>
    with _BlocStreamProviderMixin<Bloc, Value> {
  BlocStreamProvider(this._create, {String? name}) : super(name);

  static const family = BlocStreamProviderFamilyBuilder();

  static const autoDispose = AutoDisposeBlocStreamProviderBuilder();

  final Create<Bloc, ProviderReference> _create;

  @override
  BlocStreamProviderFamily<Bloc, Value, Object?>? get from =>
      super.from as BlocStreamProviderFamily<Bloc, Value, Object?>?;

  /// {@template riverpod.blocstreamprovider.blocStream}
  /// Obtains the [Bloc] associated with this [BlocStreamProvider],
  /// without listening to it.
  ///
  /// Listening to this provider may cause providers/widgets to rebuild in the
  /// event that the [Bloc] it recreated.
  /// {@endtemplate}
  @override
  late final AlwaysAliveProviderBase<Bloc, Bloc> bloc = from != null
      ? from!._blocFamily(argument)
      : _BlocStreamProvider(_create, name: name);

  /// Overrides the behavior of a provider with a another provider.
  ///
  /// {@macro riverpod.overideWith}
  Override overrideWithProvider(
    BlocStreamProvider<Bloc, Value> provider,
  ) {
    return ProviderOverride(provider.bloc, bloc);
  }

  @override
  Bloc create(ProviderReference ref) => ref.watch(bloc);

  @override
  _BlocStreamProviderState<Bloc, Value> createState() =>
      _BlocStreamProviderState<Bloc, Value>();
}

// ignore: subtype_of_sealed_class
@sealed
class BlocStreamProviderFamily<Bloc extends BlocStream<Value>, Value, Param>
    extends Family<Bloc, Value, Param, ProviderReference,
        BlocStreamProvider<Bloc, Value>> {
  BlocStreamProviderFamily(
    Bloc Function(ProviderReference ref, Param a) create, {
    String? name,
  }) : super(create, name);

  @override
  BlocStreamProvider<Bloc, Value> create(
    Param value,
    Bloc Function(ProviderReference ref, Param param) builder,
    String? name,
  ) {
    return BlocStreamProvider((ref) => builder(ref, value), name: name);
  }

  late final _blocFamily = _BlocStreamFamily<Bloc, Param>(builder, name);

  Override overrideWithProvider(
    Bloc Function(ProviderReference ref, Param param) builderOverride,
  ) {
    return FamilyOverride(
      _blocFamily,
      (param) {
        return _blocFamily.create(param as Param, builderOverride, null);
      },
    );
  }
}

// ignore: subtype_of_sealed_class
@sealed
class _BlocStreamFamily<Bloc extends BlocStream<Object?>, Param> extends Family<
    Bloc, Bloc, Param, ProviderReference, _BlocStreamProvider<Bloc>> {
  _BlocStreamFamily(
    Bloc Function(ProviderReference ref, Param param) builder,
    String? name,
  ) : super(builder, name);

  @override
  _BlocStreamProvider<Bloc> create(
    Param value,
    Bloc Function(ProviderReference ref, Param param) builder,
    String? name,
  ) {
    return _BlocStreamProvider((ref) => builder(ref, value), name: name);
  }
}

// ignore: subtype_of_sealed_class
class _AutoDisposeBlocStreamProvider<Bloc extends BlocStream<Object?>>
    extends AutoDisposeProvider<Bloc> {
  _AutoDisposeBlocStreamProvider(
    Create<Bloc, AutoDisposeProviderReference> create, {
    required String? name,
  }) : super(
          (ref) {
            final bloc = create(ref);
            ref.onDispose(bloc.close);
            return bloc;
          },
          name: name,
        );
}

@sealed
class AutoDisposeBlocStreamProvider<Bloc extends BlocStream<Value>, Value>
    extends AutoDisposeProviderBase<Bloc, Value>
    with _BlocStreamProviderMixin<Bloc, Value> {
  AutoDisposeBlocStreamProvider(this._create, {String? name}) : super(name);

  static const family = AutoDisposeBlocStreamProviderFamilyBuilder();

  static const autoDispose = AutoDisposeBlocStreamProviderBuilder();

  final Create<Bloc, AutoDisposeProviderReference> _create;

  @override
  AutoDisposeBlocStreamProviderFamily<Bloc, Value, Object?>? get from =>
      super.from as AutoDisposeBlocStreamProviderFamily<Bloc, Value, Object?>?;

  @override
  late final AutoDisposeProviderBase<Bloc, Bloc> bloc = from != null
      ? from!._blocFamily(argument)
      : _AutoDisposeBlocStreamProvider(_create, name: name);

  Override overrideWithProvider(
    AutoDisposeBlocStreamProvider<Bloc, Value> provider,
  ) {
    return ProviderOverride(provider.bloc, bloc);
  }

  @override
  Bloc create(covariant AutoDisposeProviderReference ref) {
    return ref.watch(bloc);
  }

  @override
  _BlocStreamProviderState<Bloc, Value> createState() {
    return _BlocStreamProviderState<Bloc, Value>();
  }
}

@sealed
class AutoDisposeBlocStreamProviderFamily<Notifier extends BlocStream<Value>,
        Value, Param>
    extends Family<Notifier, Value, Param, AutoDisposeProviderReference,
        AutoDisposeBlocStreamProvider<Notifier, Value>> {
  AutoDisposeBlocStreamProviderFamily(
    Notifier Function(AutoDisposeProviderReference ref, Param a) create, {
    String? name,
  }) : super(create, name);

  @override
  AutoDisposeBlocStreamProvider<Notifier, Value> create(
    Param value,
    Notifier Function(AutoDisposeProviderReference ref, Param param) builder,
    String? name,
  ) {
    return AutoDisposeBlocStreamProvider(
      (ref) => builder(ref, value),
      name: name,
    );
  }

  late final _blocFamily =
      _AutoDisposeBlocStreamFamily<Notifier, Param>(builder, name);

  /// Overrides the behavior of a family for a part of the application.
  Override overrideWithProvider(
    Notifier Function(ProviderReference ref, Param param) builderOverride,
  ) {
    return FamilyOverride(
      _blocFamily,
      (param) {
        return _blocFamily.create(param as Param, builderOverride, null);
      },
    );
  }
}

@sealed
class _AutoDisposeBlocStreamFamily<Bloc extends BlocStream<Object?>, Param>
    extends Family<Bloc, Bloc, Param, AutoDisposeProviderReference,
        _AutoDisposeBlocStreamProvider<Bloc>> {
  _AutoDisposeBlocStreamFamily(
    Bloc Function(AutoDisposeProviderReference ref, Param param) builder,
    String? name,
  ) : super(builder, name);

  @override
  _AutoDisposeBlocStreamProvider<Bloc> create(
    Param value,
    Bloc Function(AutoDisposeProviderReference ref, Param param) builder,
    String? name,
  ) {
    return _AutoDisposeBlocStreamProvider(
      (ref) => builder(ref, value),
      name: name,
    );
  }
}
